# is_matrix_forge.led_matrix.display.grid.grid
"""
Grid module for LED Matrix display.

Author: 
    Inspyre Softworks

Project:
    led-matrix-battery

File: 
    is_matrix_forge/led_matrix/display/grid/grid.py

Description:
    This module provides the Grid class which represents a 2D grid of pixels
    for the LED matrix display. It handles grid creation, manipulation, and
    loading from files.
"""

import itertools
from pathlib import Path
from typing import List, Optional, Union, ClassVar, Type, Any, Dict  # Added Any, Dict
from ...constants import WIDTH as __WIDTH, HEIGHT as __HEIGHT, PRESETS_DIR
from .helpers import is_valid_grid, generate_blank_grid
from is_matrix_forge.common.helpers import coerce_to_int

MATRIX_HEIGHT = __HEIGHT
"""int: Height of the LED matrix grid in pixels.
This constant defines the default height for new Grid instances and for loading operations.
"""

MATRIX_WIDTH = __WIDTH
"""int: Width of the LED matrix grid in pixels.
This constant defines the default width for new Grid instances and for loading operations.
"""


class Grid:
    """
    Represents a 2D column-major grid for the LED display (grid[x][y], 9×34).
    Provides methods for grid creation, manipulation, and loading from files.

    The Grid class allows for the management of pixel data in a column-major format,
    supporting operations such as shifting, drawing, and loading from specifications or files.
    """

    def __init__(
        self,
        width: int = MATRIX_WIDTH,
        height: int = MATRIX_HEIGHT,
        fill_value: int = 0,
        init_grid: List[List[int]] = None
    ) -> None:
        """
        Initialize a Grid. If `init_grid` is provided, it must be column-major
        with shape (width × height). Otherwise, create a blank grid.
        """
        if init_grid is not None:
            if not is_valid_grid(init_grid, width, height):
                raise ValueError(f"init_grid must be {width}×{height} column-major 0/1 list")
            self._grid = [col[:] for col in init_grid]
        else:
            if fill_value not in (0, 1):
                raise ValueError("fill_value must be 0 or 1")
            self._grid = generate_blank_grid(width=width, height=height, fill_value=fill_value)

        self._width = width
        self._height = height
        self._fill_value = fill_value

    @property
    def grid(self) -> List[List[int]]:
        """Defensive copy of the column-major grid data."""
        return [col[:] for col in self._grid]

    @grid.setter
    def grid(self, value: List[List[int]]) -> None:
        """Replace the internal grid; must be column-major and correct shape."""
        if not is_valid_grid(value, self._width, self._height):
            raise ValueError(f"grid must be {self._width}×{self._height} column-major 0/1 list")
        self._grid = [col[:] for col in value]

    @property
    def width(self) -> int:
        """Number of columns."""
        return self._width

    @property
    def height(self) -> int:
        """Number of rows."""
        return self._height

    @property
    def fill_value(self) -> int:
        """Default pixel fill (0 or 1)."""
        return self._fill_value

    @fill_value.setter
    def fill_value(self, v: int) -> None:
        """Set default pixel fill; must be 0 or 1."""
        if not isinstance(v, int) or v not in (0, 1):
            raise ValueError("fill_value must be 0 or 1")
        self._fill_value = v

    @property
    def cols(self) -> int:
        """Alias for width."""
        return self._width

    @property
    def rows(self) -> int:
        """Alias for height."""
        return self._height

    @classmethod
    def load_blank_grid(
        cls,
        width:      int = MATRIX_WIDTH,
        height:     int = MATRIX_HEIGHT,
        fill_value: int = 0
    ) -> List[List[int]]:
        """Return a new blank column-major grid."""
        return generate_blank_grid(width=width, height=height, fill_value=fill_value)

    @classmethod
    def from_spec(
        cls,
        spec: List[List[int]]
    ) -> 'Grid':
        """Instantiate directly from a column-major spec list."""
        return cls(init_grid=spec)

    @classmethod
    def from_file(
        cls,
        filename: Union[str, Path],
        frame_number: int = 0,
        height: int = MATRIX_HEIGHT,
        width: int = MATRIX_WIDTH
    ) -> 'Grid':
        """
        Load a column-major grid from file (single grid or frames of grids).
        """
        from ...helpers import load_from_file
        raw = load_from_file(str(filename))
        # Single-grid JSON: list of lists
        if isinstance(raw, list) and raw and isinstance(raw[0], list):
            grid_data = raw
        # Frame-list JSON: list of dicts
        elif isinstance(raw, list) and all(isinstance(f, dict) for f in raw):
            frame = raw[frame_number]
            grid_data = frame.get('grid')
            if not isinstance(grid_data, list):
                raise ValueError(f"Frame {frame_number} missing 'grid' list")
        else:
            raise ValueError("Unsupported file structure for grid data.")

        if not is_valid_grid(grid_data, width, height):
            raise ValueError(f"Loaded grid is not {width}×{height} column-major 0/1 list")

        return cls(width=width, height=height, init_grid=grid_data)

    def draw(self, device: Any) -> None:
        """Draw this grid via device.draw_grid(grid)."""
        if not hasattr(device, 'draw_grid') or not callable(device.draw_grid):
            raise AttributeError("device.draw_grid(grid) not available")
        device.draw_grid(self)

    def get_pixel_value(self, x: int, y: int) -> int:
        """Return the value at column x, row y."""
        if x < 0 or x >= self._width or y < 0 or y >= self._height:
            raise IndexError(f"({x},{y}) out of bounds {self._width}×{self._height}")
        return self._grid[x][y]

    def get_shifted(
        self,
        dx: int = 0,
        dy: int = 0,
        wrap: bool = False
    ) -> 'Grid':
        """
        Return a new Grid shifted by (dx, dy):
          - dx > 0 moves right, dx < 0 moves left
          - dy > 0 moves down, dy < 0 moves up
        If wrap=True, shifts wrap around edges; otherwise, out-of-bounds fill with fill_value.
        """
        new = generate_blank_grid(self._width, self._height, self._fill_value)

        for c, r in itertools.product(range(self._width), range(self._height)):
            src_c = (c + dx) % self._width if wrap else c + dx
            src_r = (r + dy) % self._height if wrap else r + dy

            if 0 <= src_c < self._width and 0 <= src_r < self._height:
                new[c][r] = self._grid[src_c][src_r]

        return Grid(width=self._width, height=self._height, fill_value=self._fill_value, init_grid=new)

    def __getitem__(self, index: int) -> list[int]:
        """Allow column-major indexing: grid[col] → list of pixel values down that column."""
        return self._grid[index]

    def __len__(self) -> int:
        """Number of columns (i.e. grid width)."""
        return len(self._grid)

    def __iter__(self):
        """Iterate over columns."""
        return iter(self._grid)
